var falcor = require("./../../lib/");
var Model = falcor.Model;
var expect = require('chai').expect;
var $path = require("./../../lib/types/ref");
var $atom = require("./../../lib/types/atom");
var testRunner = require('../testRunner');
var sinon = require('sinon');
var noOp = function() {};
var Rx = require('rx');
var Observable = Rx.Observable;

describe("Special Cases", function() {
    it('should set in an array and the length should be set in.', function(done) {
        var model = new Model();
        var onNext = sinon.spy();
        model.
            set({
                json: {
                    foo: ['bar']
                }
            }).
            flatMap(function() {
                return model.get('foo.length');
            }).
            doAction(onNext).
            doAction(noOp, noOp, function() {
                expect(onNext.calledOnce).to.be.ok;
                expect(onNext.getCall(0).args[0]).to.deep.equals({
                    json: {foo: { length: 1 } }
                });
            }).
            subscribe(noOp, done, done);
    });
    it('should set the cache in.', function() {
        var model = new Model({cache: {}});
        var edgeCaseCache = {
            jsonGraph: {
                user: {
                    name: "Jim",
                    location: {$type: "error", value: "Something broke!"},
                    age: {$type: $atom}
                }
            },
            paths: [
                ['user', ['name', 'location', 'age']]
            ]
        };

        var pathMap = [{}];
        model._setJSONGsAsPathMap(model, [edgeCaseCache], pathMap);

        testRunner.compare({
            json: {
                user: {
                    name: 'Jim'
                }
            }
        }, pathMap[0]);

        var jsons = [{}];
        model._root.cache = {};
        model._setJSONGsAsJSON(model, [edgeCaseCache], jsons);
        testRunner.compare({ json: { name: 'Jim' } }, jsons[0]);
    });
    it("set blows away the cache.", function() {
        var model = new Model({});
        var get = [["genreList", 1, 0, "summary"]];

        // this mimicks the server setting cycle from the router.
        var set = [
            {
                jsonGraph: {"genreList": {
                    "0": { "$type": $path, "value": ["lists", "abcd"] },
                    "1": { "$type": $path, "value": ["lists", "my-list"] }
                }},
                paths: [['genreList', {to:1}, 0, 'summary']]
            },
            {
                jsonGraph: {"lists": {
                    "abcd": { "0": { "$type": $path, "value": ["videos", 1234] } },
                    "my-list": { "$type": $path, "value": ["lists", "1x5x"] }
                }},
                paths: [["genreList", 1, 0, "summary"]]
            },
            {
                jsonGraph: {"lists": {"1x5x": {
                    "0": { "$type": $path, "value": ["videos", 553] }
                }}},
                paths: [["genreList", 1, 0, "summary"]]
            },
            {
                jsonGraph: {"videos": {"553": {"summary": {
                    "$size": 10,
                    "$type": $atom,
                    "value": {
                        "title": "Running Man",
                        "url": "/movies/553"
                    }
                }}}},
                paths: [["genreList", 1, 0, "summary"]]
            }
        ];

        var seed = [{}];
        set.forEach(function(s, i) {
            model._setJSONGsAsPathMap(model, [s], seed);
            if (i === 2) {
                expect(model._root.cache.lists).to.be.ok;
            }
        });

        model._getPathValuesAsValues(model, get, function(x) {
            expect(x).to.deep.equals({
                value: {
                    "title": "Running Man",
                    "url": "/movies/553"
                },
                path: ["genreList", 1, 0, "summary"]
            });
        });
    });

    it('should return the ranged items when ranges in array.', function() {
        var JSONG = {
            jsonGraph: {
                foo: {
                    0: {
                        $type: $atom,
                        value: 0
                    },
                    1: {
                        $type: $atom,
                        value: 75
                    }
                }
            },
            paths: [
                ['foo', [{from: 0, to: 1}]]
            ]
        };
        var model = new Model();
        var out = [{}];

        model._setJSONGsAsJSON(model, [JSONG], out);
        expect(out[0].json).to.deep.equals({
            0: 0,
            1: 75
        });
    });
});

